;; Generated package description from org-auto-tangle.el  -*- no-byte-compile: t -*-
(define-package "org-auto-tangle" "0.6.0.0.20251230.141613" "Automatically and Asynchronously tangles org files on save" '((emacs "24.1") (async "1.9.3")) :commit "b4e7abc019179df473ecddf0af80561ddad8fc58" :authors '(("Yilkal Argaw" . "yilkalargawworkneh@gmail.com")) :maintainer '("Yilkal Argaw" . "yilkalargawworkneh@gmail.com") :keywords '("outlines") :url "https://github.com/yilkalargaw/org-auto-tangle")
